# Copyright (C) 2020-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import datetime, timedelta
import json
from unittest.mock import Mock

from django.conf import settings
from django.contrib.auth import authenticate, get_backends
from django.core.cache import cache
import pytest
from rest_framework.exceptions import AuthenticationFailed

from swh.auth.django.backends import OIDCBearerTokenAuthentication
from swh.auth.django.models import OIDCUser
from swh.auth.django.utils import oidc_profile_cache_key, reverse
from swh.auth.keycloak import ExpiredSignatureError, KeycloakError

pytestmark = pytest.mark.django_db


def _authenticate_user(request_factory):
    request = request_factory.get(reverse("root"))

    return authenticate(
        request=request,
        code="some-code",
        code_verifier="some-code-verifier",
        redirect_uri="https://localhost:5004",
    )


def _check_authenticated_user(user, decoded_token, keycloak_oidc):
    assert user is not None
    assert isinstance(user, OIDCUser)
    assert user.id != 0
    assert user.username == decoded_token["preferred_username"]
    assert user.password == ""
    assert user.first_name == decoded_token["given_name"]
    assert user.last_name == decoded_token["family_name"]
    assert user.email == decoded_token["email"]
    assert user.is_staff == ("/staff" in decoded_token["groups"])
    assert {group.name for group in user.groups.all()} == {
        group_name.lstrip("/") for group_name in keycloak_oidc.user_groups
    }
    assert user.sub == decoded_token["sub"]
    resource_access = decoded_token.get("resource_access", {})
    resource_access_client = resource_access.get(keycloak_oidc.client_id, {})
    assert user.permissions == set(resource_access_client.get("roles", []))
    assert all(user.has_perm(perm) for perm in resource_access_client.get("roles", []))


def test_oidc_code_pkce_auth_backend_success(keycloak_oidc, request_factory):
    """
    Checks successful login based on OpenID Connect with PKCE extension
    Django authentication backend (login from Web UI).
    """
    keycloak_oidc.user_groups = ["/staff", "/other_group"]

    oidc_profile = keycloak_oidc.login()
    user = _authenticate_user(request_factory)

    decoded_token = keycloak_oidc.decode_token(user.access_token)
    _check_authenticated_user(user, decoded_token, keycloak_oidc)

    auth_datetime = datetime.fromtimestamp(decoded_token["iat"])
    exp_datetime = datetime.fromtimestamp(decoded_token["exp"])
    refresh_exp_datetime = auth_datetime + timedelta(
        seconds=oidc_profile["refresh_expires_in"]
    )

    assert user.access_token == oidc_profile["access_token"]
    assert user.expires_at == exp_datetime
    assert user.id_token == oidc_profile["id_token"]
    assert user.refresh_token == oidc_profile["refresh_token"]
    assert user.refresh_expires_at == refresh_exp_datetime
    assert user.scope == oidc_profile["scope"]
    assert user.session_state == oidc_profile["session_state"]

    backend_path = "swh.auth.django.backends.OIDCAuthorizationCodePKCEBackend"
    assert user.backend == backend_path
    backend_idx = settings.AUTHENTICATION_BACKENDS.index(backend_path)
    assert get_backends()[backend_idx].get_user(user.id) == user


def test_oidc_code_pkce_auth_backend_failure(keycloak_oidc, request_factory):
    """
    Checks failed login based on OpenID Connect with PKCE extension Django
    authentication backend (login from Web UI).
    """
    keycloak_oidc.set_auth_success(False)

    user = _authenticate_user(request_factory)

    assert user is None


def test_oidc_code_pkce_auth_backend_refresh_token_success(
    keycloak_oidc, request_factory
):
    """
    Checks access token renewal success using refresh token.
    """

    oidc_profile = keycloak_oidc.login()
    decoded_token = keycloak_oidc.decode_token(oidc_profile["access_token"])

    keycloak_oidc.decode_token = Mock()
    keycloak_oidc.decode_token.side_effect = [
        ExpiredSignatureError("access token token has expired"),
        decoded_token,
    ]

    user = _authenticate_user(request_factory)

    oidc_profile = keycloak_oidc.login()
    keycloak_oidc.refresh_token.assert_called_with(oidc_profile["refresh_token"])

    assert user is not None


def test_oidc_code_pkce_auth_backend_refresh_token_failure(
    keycloak_oidc, request_factory
):
    """
    Checks access token renewal failure using refresh token.
    """

    # authenticate user
    user = _authenticate_user(request_factory)
    assert user is not None
    # OIDC profile should be in cache
    cache_key = oidc_profile_cache_key(keycloak_oidc, user.id)
    assert cache.get(cache_key) is not None

    # simulate terminated OIDC session
    keycloak_oidc.decode_token = Mock()
    keycloak_oidc.decode_token.side_effect = ExpiredSignatureError(
        "access token token has expired"
    )

    kc_error_dict = {
        "error": "invalid_grant",
        "error_description": "Session not active",
    }
    keycloak_oidc.refresh_token.side_effect = KeycloakError(
        error_message=json.dumps(kc_error_dict).encode(), response_code=400
    )

    backend_path = "swh.auth.django.backends.OIDCAuthorizationCodePKCEBackend"
    assert user.backend == backend_path
    backend_idx = settings.AUTHENTICATION_BACKENDS.index(backend_path)

    # try to authenticate user again from its id and cached OIDC profile
    user = get_backends()[backend_idx].get_user(user.id)

    # it should have tried to refresh token
    oidc_profile = keycloak_oidc.login()
    keycloak_oidc.refresh_token.assert_called_with(oidc_profile["refresh_token"])

    # authentication failed
    assert user is None
    # invalid OIDC profile should have been removed from cache
    assert cache.get(cache_key) is None


def test_oidc_code_pkce_auth_backend_permissions(keycloak_oidc, request_factory):
    """
    Checks that a permission defined with OpenID Connect is correctly mapped
    to a Django one when logging from Web UI.
    """
    realm_permission = "swh.some-permission"
    client_permission = "webapp.some-permission"
    keycloak_oidc.realm_permissions = [realm_permission]
    keycloak_oidc.client_permissions = [client_permission]
    user = _authenticate_user(request_factory)
    assert user.has_perm(realm_permission)
    assert user.has_perm(client_permission)
    assert user.get_all_permissions() == {realm_permission, client_permission}
    assert user.get_group_permissions() == {realm_permission, client_permission}
    assert user.has_module_perms("webapp")
    assert not user.has_module_perms("foo")


def test_drf_oidc_bearer_token_auth_backend_success(keycloak_oidc, api_request_factory):
    """
    Checks successful login based on OpenID Connect bearer token Django REST
    Framework authentication backend (Web API login).
    """
    url = reverse("api-test")
    drf_auth_backend = OIDCBearerTokenAuthentication()

    oidc_profile = keycloak_oidc.login()
    refresh_token = oidc_profile["refresh_token"]
    access_token = oidc_profile["access_token"]

    decoded_token = keycloak_oidc.decode_token(access_token)

    request = api_request_factory.get(url, HTTP_AUTHORIZATION=f"Bearer {refresh_token}")

    user, _ = drf_auth_backend.authenticate(request)
    _check_authenticated_user(user, decoded_token, keycloak_oidc)
    # oidc_profile is not filled when authenticating through bearer token
    assert hasattr(user, "access_token") and user.access_token is None


def test_drf_oidc_bearer_token_auth_backend_failure(keycloak_oidc, api_request_factory):
    """
    Checks failed login based on OpenID Connect bearer token Django REST
    Framework authentication backend (Web API login).
    """
    url = reverse("api-test")
    drf_auth_backend = OIDCBearerTokenAuthentication()

    oidc_profile = keycloak_oidc.login()

    # simulate a failed authentication with a bearer token in expected format
    keycloak_oidc.set_auth_success(False)

    refresh_token = oidc_profile["refresh_token"]

    request = api_request_factory.get(url, HTTP_AUTHORIZATION=f"Bearer {refresh_token}")

    with pytest.raises(AuthenticationFailed):
        drf_auth_backend.authenticate(request)

    # simulate a failed authentication with an invalid bearer token format
    request = api_request_factory.get(
        url, HTTP_AUTHORIZATION="Bearer invalid-token-format"
    )

    with pytest.raises(AuthenticationFailed):
        drf_auth_backend.authenticate(request)


def test_drf_oidc_auth_invalid_or_missing_auth_type(keycloak_oidc, api_request_factory):
    """
    Checks failed login based on OpenID Connect bearer token Django REST
    Framework authentication backend (Web API login) due to invalid
    authorization header value.
    """
    url = reverse("api-test")
    drf_auth_backend = OIDCBearerTokenAuthentication()

    oidc_profile = keycloak_oidc.login()
    refresh_token = oidc_profile["refresh_token"]

    # Invalid authorization type
    request = api_request_factory.get(url, HTTP_AUTHORIZATION="Foo token")

    with pytest.raises(AuthenticationFailed):
        drf_auth_backend.authenticate(request)

    # Missing authorization type
    request = api_request_factory.get(url, HTTP_AUTHORIZATION=f"{refresh_token}")

    with pytest.raises(AuthenticationFailed):
        drf_auth_backend.authenticate(request)


def test_drf_oidc_bearer_token_auth_backend_permissions(
    keycloak_oidc, api_request_factory
):
    """
    Checks that a permission defined with OpenID Connect is correctly mapped
    to a Django one when using bearer token authentication.
    """
    realm_permission = "swh.some-permission"
    client_permission = "webapp.some-permission"
    keycloak_oidc.realm_permissions = [realm_permission]
    keycloak_oidc.client_permissions = [client_permission]

    drf_auth_backend = OIDCBearerTokenAuthentication()
    oidc_profile = keycloak_oidc.login()
    refresh_token = oidc_profile["refresh_token"]
    url = reverse("api-test")
    request = api_request_factory.get(url, HTTP_AUTHORIZATION=f"Bearer {refresh_token}")
    user, _ = drf_auth_backend.authenticate(request)

    assert user.has_perm(realm_permission)
    assert user.has_perm(client_permission)
    assert user.get_all_permissions() == {realm_permission, client_permission}
    assert user.get_group_permissions() == {realm_permission, client_permission}
    assert user.has_module_perms("webapp")
    assert not user.has_module_perms("foo")



# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from django.conf.urls import url
from django.contrib.auth.views import LogoutView
from django.http import HttpResponse
from rest_framework.decorators import api_view
from rest_framework.response import Response

from swh.auth.django.views import urlpatterns as auth_urlpatterns


def _root_view(request):
    return HttpResponse("Hello World !")


@api_view()
def _api_view_test(request):
    return Response({"message": "Hello World !"})


urlpatterns = [
    url(r"^$", _root_view, name="root"),
    url(r"^api/test$", _api_view_test, name="api-test"),
    url(r"^logout/$", LogoutView.as_view(), name="logout"),
] + auth_urlpatterns



# Copyright (C) 2020-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from django.conf import settings
from django.contrib.auth import BACKEND_SESSION_KEY
from django.http.response import HttpResponseRedirect

from swh.auth.django.utils import reverse


class OIDCSessionExpiredMiddleware:
    """
    Middleware for checking OpenID Connect user session expiration.

    That middleware detects when a user previously logged in using
    the OpenID Connect authentication backend got his session expired.

    In that case it will perform a redirection to a django view whose
    name must be set in the ``SWH_AUTH_SESSION_EXPIRED_REDIRECT_VIEW``
    django setting (typically a logout view).

    The following query parameter will be set for that view:

      * ``next_path``: requested URL before the detection of the session expiration
      * ``remote_user``: indicates that the user was previously authenticated with OIDC

    """

    def __init__(self, get_response=None):
        self.get_response = get_response
        self.redirect_view = getattr(
            settings, "SWH_AUTH_SESSION_EXPIRED_REDIRECT_VIEW", None
        )
        if self.redirect_view is None:
            raise ValueError(
                "SWH_AUTH_SESSION_EXPIRED_REDIRECT_VIEW django setting "
                "is mandatory to instantiate OIDCSessionExpiredMiddleware class"
            )
        self.exempted_urls = [
            reverse(v)
            for v in (
                self.redirect_view,
                "oidc-login",
                "oidc-login-complete",
                "oidc-logout",
            )
        ]

    def __call__(self, request):
        if (
            request.method != "GET"
            or request.user.is_authenticated
            or BACKEND_SESSION_KEY not in request.session
            or "OIDC" not in request.session[BACKEND_SESSION_KEY]
            or request.path in self.exempted_urls
        ):
            return self.get_response(request)

        # At that point, we know that a OIDC user was previously logged in
        # and his session has expired.
        # Redirect to a view specified in django settings.
        next_path = request.get_full_path()
        logout_url = reverse(
            self.redirect_view, query_params={"next_path": next_path, "remote_user": 1}
        )
        return HttpResponseRedirect(logout_url)



# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Set

import pytest

from swh.auth.django.models import OIDCUser

PERMISSIONS: Set[str] = set(["api", "app-label-read"])
NO_PERMISSION: Set[str] = set()


@pytest.fixture
def oidc_user():
    return OIDCUser(
        id=666,
        username="foo",
        password="bar",
        first_name="foobar",
        last_name="",
        email="foo@bar.org",
    )


@pytest.fixture
def oidc_user_admin(oidc_user):
    oidc_user_admin = oidc_user
    oidc_user_admin.is_active = True
    oidc_user_admin.is_superuser = True
    return oidc_user_admin


def test_django_oidc_user(oidc_user):
    oidc_user.permissions = PERMISSIONS

    assert oidc_user.get_group_permissions() == PERMISSIONS
    assert oidc_user.get_group_permissions(oidc_user) == PERMISSIONS
    assert oidc_user.get_all_permissions() == PERMISSIONS
    assert oidc_user.get_all_permissions(oidc_user) == PERMISSIONS

    assert "api" in PERMISSIONS
    assert oidc_user.has_perm("api") is True
    assert oidc_user.has_perm("something") is False

    assert "app-label-read" in PERMISSIONS
    assert oidc_user.has_module_perms("app-label") is True
    assert oidc_user.has_module_perms("app-something") is False


def test_django_oidc_user_admin(oidc_user_admin):
    oidc_user_admin.permissions = NO_PERMISSION

    assert oidc_user_admin.get_group_permissions() == NO_PERMISSION
    assert oidc_user_admin.get_group_permissions(oidc_user_admin) == NO_PERMISSION

    assert oidc_user_admin.get_all_permissions() == NO_PERMISSION
    assert oidc_user_admin.get_all_permissions(oidc_user) == NO_PERMISSION

    assert "foobar" not in PERMISSIONS
    assert oidc_user_admin.has_perm("foobar") is True
    assert "something" not in PERMISSIONS
    assert oidc_user_admin.has_perm("something") is True

    assert oidc_user_admin.has_module_perms("app-label") is True
    assert oidc_user_admin.has_module_perms("really-whatever-app") is True



# Copyright (C) 2021-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from copy import copy
from datetime import datetime

from django.test import override_settings
import pytest

from swh.auth.django.utils import (
    keycloak_oidc_client,
    oidc_user_from_decoded_token,
    oidc_user_from_profile,
)
from swh.auth.tests.sample_data import (
    CLIENT_ID,
    DECODED_TOKEN,
    OIDC_PROFILE,
    REALM_NAME,
    SERVER_URL,
)

pytestmark = pytest.mark.django_db


def _check_user(user, is_staff=False, permissions=set(), groups=set()):
    assert user.id > 0
    assert user.username == DECODED_TOKEN["preferred_username"]
    assert user.password == ""
    assert user.first_name == DECODED_TOKEN["given_name"]
    assert user.last_name == DECODED_TOKEN["family_name"]
    assert user.email == DECODED_TOKEN["email"]
    assert user.is_staff == is_staff
    assert {group.name for group in user.groups.all()} == groups
    assert user.permissions == permissions
    assert all(user.has_perm(perm) for perm in permissions)
    assert user.sub == DECODED_TOKEN["sub"]

    date_now = datetime.now()
    if user.expires_at is not None:
        assert isinstance(user.expires_at, datetime)
        assert date_now <= user.expires_at
    if user.refresh_expires_at is not None:
        assert isinstance(user.refresh_expires_at, datetime)
        assert date_now <= user.refresh_expires_at

    assert user.oidc_profile == {
        k: getattr(user, k)
        for k in (
            "access_token",
            "expires_in",
            "expires_at",
            "id_token",
            "refresh_token",
            "refresh_expires_in",
            "refresh_expires_at",
            "scope",
            "session_state",
        )
    }


def test_oidc_user_from_decoded_token():
    user = oidc_user_from_decoded_token(DECODED_TOKEN)
    _check_user(user)


def test_oidc_user_with_permissions_from_decoded_token():
    decoded_token = copy(DECODED_TOKEN)
    decoded_token["groups"] = ["/staff", "api"]
    decoded_token["realm_access"] = {"roles": ["swh.ambassador"]}
    decoded_token["resource_access"] = {CLIENT_ID: {"roles": ["read-api"]}}

    user = oidc_user_from_decoded_token(decoded_token, client_id=CLIENT_ID)

    _check_user(
        user,
        is_staff=True,
        permissions={"swh.ambassador", "read-api"},
        groups={group_name.lstrip("/") for group_name in decoded_token["groups"]},
    )


@pytest.mark.parametrize(
    "key,mapped_key",
    [
        ("preferred_username", "username"),
        ("given_name", "first_name"),
        ("family_name", "last_name"),
        ("email", "email"),
    ],
)
def test_oidc_user_from_decoded_token_empty_fields_ok(key, mapped_key):
    decoded_token = copy(DECODED_TOKEN)
    decoded_token.pop(key, None)

    user = oidc_user_from_decoded_token(decoded_token, client_id=CLIENT_ID)

    # Ensure the missing field is mapped to an empty value
    assert getattr(user, mapped_key) == ""


def test_oidc_user_from_profile(keycloak_oidc):
    user = oidc_user_from_profile(keycloak_oidc, OIDC_PROFILE)
    _check_user(user)


@override_settings(
    SWH_AUTH_SERVER_URL=None,
    SWH_AUTH_REALM_NAME=None,
    SWH_AUTH_CLIENT_ID=None,
)
def test_keycloak_oidc_client_missing_django_settings():

    with pytest.raises(ValueError, match="settings are mandatory"):
        keycloak_oidc_client()


@override_settings(
    SWH_AUTH_SERVER_URL=SERVER_URL,
    SWH_AUTH_REALM_NAME=REALM_NAME,
    SWH_AUTH_CLIENT_ID=CLIENT_ID,
)
def test_keycloak_oidc_client_parameters_from_django_settings():

    kc_oidc_client = keycloak_oidc_client()

    assert kc_oidc_client.server_url == SERVER_URL
    assert kc_oidc_client.realm_name == REALM_NAME
    assert kc_oidc_client.client_id == CLIENT_ID



# Copyright (C) 2020-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json

from django.contrib.auth.models import AnonymousUser, User
from django.core.cache import cache
import pytest

from swh.auth.django.models import OIDCUser
from swh.auth.django.utils import reverse
from swh.auth.keycloak import KeycloakError


@pytest.fixture
def api_client(api_client):
    # ensure django cache is cleared before each test
    cache.clear()
    return api_client


@pytest.mark.django_db
def test_drf_django_session_auth_success(keycloak_oidc, client):
    """
    Check user gets authenticated when querying the web api
    through a web browser.
    """
    url = reverse("api-test")

    client.login(code="", code_verifier="", redirect_uri="")

    response = client.get(url)
    assert response.status_code == 200
    request = response.wsgi_request

    # user should be authenticated
    assert isinstance(request.user, OIDCUser)

    # check remoter used has not been saved to Django database
    with pytest.raises(User.DoesNotExist):
        User.objects.get(username=request.user.username)


@pytest.mark.django_db
def test_drf_oidc_bearer_token_auth_success(keycloak_oidc, api_client):
    """
    Check user gets authenticated when querying the web api
    through an HTTP client using bearer token authentication.
    """
    url = reverse("api-test")

    oidc_profile = keycloak_oidc.login()
    refresh_token = oidc_profile["refresh_token"]

    api_client.credentials(HTTP_AUTHORIZATION=f"Bearer {refresh_token}")

    response = api_client.get(url)
    assert response.status_code == 200
    request = response.wsgi_request

    # user should be authenticated
    assert isinstance(request.user, OIDCUser)

    # check remoter used has not been saved to Django database
    with pytest.raises(User.DoesNotExist):
        User.objects.get(username=request.user.username)


@pytest.mark.django_db
def test_drf_oidc_bearer_token_auth_failure(keycloak_oidc, api_client):
    url = reverse("api-test")

    oidc_profile = keycloak_oidc.login()
    refresh_token = oidc_profile["refresh_token"]

    # check for failed authentication but with expected token format
    keycloak_oidc.set_auth_success(False)
    api_client.credentials(HTTP_AUTHORIZATION=f"Bearer {refresh_token}")

    response = api_client.get(url)
    assert response.status_code == 403
    request = response.wsgi_request

    assert isinstance(request.user, AnonymousUser)

    # check for failed authentication when token format is invalid
    api_client.credentials(HTTP_AUTHORIZATION="Bearer invalid-token-format-ééàà")

    response = api_client.get(url)
    assert response.status_code == 400
    request = response.wsgi_request

    assert isinstance(request.user, AnonymousUser)


def test_drf_oidc_auth_invalid_or_missing_authorization_type(keycloak_oidc, api_client):
    url = reverse("api-test")

    oidc_profile = keycloak_oidc.login()
    refresh_token = oidc_profile["refresh_token"]

    # missing authorization type
    api_client.credentials(HTTP_AUTHORIZATION=f"{refresh_token}")

    response = api_client.get(url)
    assert response.status_code == 403
    request = response.wsgi_request

    assert isinstance(request.user, AnonymousUser)

    # invalid authorization type
    api_client.credentials(HTTP_AUTHORIZATION="Foo token")

    response = api_client.get(url)
    assert response.status_code == 403
    request = response.wsgi_request

    assert isinstance(request.user, AnonymousUser)


@pytest.mark.django_db
def test_drf_oidc_bearer_token_expired_token(keycloak_oidc, api_client):
    url = reverse("api-test")

    oidc_profile = keycloak_oidc.login()
    refresh_token = oidc_profile["refresh_token"]

    api_client.credentials(HTTP_AUTHORIZATION=f"Bearer {refresh_token}")

    for kc_err_msg in ("Offline session not active", "Offline user session not found"):

        kc_error_dict = {
            "error": "invalid_grant",
            "error_description": kc_err_msg,
        }

        keycloak_oidc.refresh_token.side_effect = KeycloakError(
            error_message=json.dumps(kc_error_dict).encode(), response_code=400
        )

        response = api_client.get(url)
        expected_error_msg = (
            "Bearer token expired after a long period of inactivity; "
            "please generate a new one."
        )

        assert response.status_code == 403
        assert expected_error_msg in json.dumps(response.data)

        request = response.wsgi_request
        assert isinstance(request.user, AnonymousUser)



# Copyright (C) 2021 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from copy import copy
import json
import os
from urllib.parse import parse_qs, urlparse

from keycloak.exceptions import KeycloakError
import pytest
import yaml

from swh.auth.keycloak import KeycloakOpenIDConnect, keycloak_error_message
from swh.auth.tests.sample_data import CLIENT_ID, DECODED_TOKEN, OIDC_PROFILE, USER_INFO
from swh.core.config import read


def test_keycloak_oidc_well_known(keycloak_oidc):
    well_known_result = keycloak_oidc.well_known()
    assert set(well_known_result.keys()) == {
        "issuer",
        "authorization_endpoint",
        "token_endpoint",
        "userinfo_endpoint",
        "end_session_endpoint",
        "jwks_uri",
        "token_introspection_endpoint",
    }


def test_keycloak_oidc_authorization_url(keycloak_oidc):
    actual_auth_uri = keycloak_oidc.authorization_url("http://redirect-uri", foo="bar")

    expected_auth_url = keycloak_oidc.well_known()["authorization_endpoint"]
    parsed_result = urlparse(actual_auth_uri)
    assert expected_auth_url.endswith(parsed_result.path)

    parsed_query = parse_qs(parsed_result.query)
    assert parsed_query == {
        "client_id": [CLIENT_ID],
        "response_type": ["code"],
        "redirect_uri": ["http://redirect-uri"],
        "foo": ["bar"],
    }


def test_keycloak_oidc_authorization_code_fail(keycloak_oidc):
    "Authorization failure raise error"
    # Simulate failed authentication with Keycloak
    keycloak_oidc.set_auth_success(False)

    with pytest.raises(KeycloakError):
        keycloak_oidc.authorization_code("auth-code", "redirect-uri")

    with pytest.raises(KeycloakError):
        keycloak_oidc.login("username", "password")


def test_keycloak_oidc_authorization_code(keycloak_oidc):
    actual_response = keycloak_oidc.authorization_code("auth-code", "redirect-uri")
    assert actual_response == OIDC_PROFILE


def test_keycloak_oidc_refresh_token(keycloak_oidc):
    actual_result = keycloak_oidc.refresh_token("refresh-token")
    assert actual_result == OIDC_PROFILE


def test_keycloak_oidc_userinfo(keycloak_oidc):
    actual_user_info = keycloak_oidc.userinfo("refresh-token")
    assert actual_user_info == USER_INFO


def test_keycloak_oidc_logout(keycloak_oidc):
    """Login out does not raise"""
    keycloak_oidc.logout("refresh-token")


def test_keycloak_oidc_decode_token(keycloak_oidc):
    actual_decoded_data = keycloak_oidc.decode_token(OIDC_PROFILE["access_token"])

    actual_decoded_data2 = copy(actual_decoded_data)
    expected_decoded_token = copy(DECODED_TOKEN)
    for dynamic_valued_key in ["exp", "iat", "auth_time"]:
        actual_decoded_data2.pop(dynamic_valued_key, None)
        expected_decoded_token.pop(dynamic_valued_key, None)

    assert actual_decoded_data2 == expected_decoded_token


def test_keycloak_oidc_login(keycloak_oidc):
    actual_response = keycloak_oidc.login("username", "password")
    assert actual_response == OIDC_PROFILE


@pytest.fixture
def auth_config():
    return {
        "keycloak": {
            "server_url": "https://auth.swh.org/SWHTest",
            "realm_name": "SWHTest",
            "client_id": "client_id",
        }
    }


@pytest.fixture
def auth_config_path(tmp_path, monkeypatch, auth_config):
    conf_path = os.path.join(tmp_path, "auth.yml")
    with open(conf_path, "w") as f:
        f.write(yaml.dump(auth_config))
    monkeypatch.setenv("SWH_CONFIG_FILENAME", conf_path)
    return conf_path


def test_auth_KeycloakOpenIDConnect_from_config(auth_config):
    """Instantiating keycloak client out of configuration dict is possible"""
    client = KeycloakOpenIDConnect.from_config(**auth_config)

    assert client.server_url == auth_config["keycloak"]["server_url"]
    assert client.realm_name == auth_config["keycloak"]["realm_name"]
    assert client.client_id == auth_config["keycloak"]["client_id"]


def test_auth_KeycloakOpenIDConnect_from_configfile(auth_config_path, monkeypatch):
    """Instantiating keycloak client out of environment variable is possible"""
    client = KeycloakOpenIDConnect.from_configfile()

    auth_config = read(auth_config_path)

    assert client.server_url == auth_config["keycloak"]["server_url"]
    assert client.realm_name == auth_config["keycloak"]["realm_name"]
    assert client.client_id == auth_config["keycloak"]["client_id"]


def test_auth_KeycloakOpenIDConnect_from_configfile_override(
    auth_config_path, monkeypatch
):
    """Instantiating keycloak client out of environment variable is possible
    And caller can override the configuration  at calling

    """
    client = KeycloakOpenIDConnect.from_configfile(client_id="foobar")

    auth_config = read(auth_config_path)

    assert client.server_url == auth_config["keycloak"]["server_url"]
    assert client.realm_name == auth_config["keycloak"]["realm_name"]
    assert client.client_id == "foobar"


@pytest.mark.parametrize(
    "error_dict, expected_result",
    [
        ({"error": "unknown_error"}, "unknown_error"),
        (
            {"error": "invalid_grant", "error_description": "Invalid credentials"},
            "invalid_grant: Invalid credentials",
        ),
    ],
)
def test_auth_keycloak_error_message(error_dict, expected_result):
    """Conversion from KeycloakError to error message should work with detail or not"""
    error_message = json.dumps(error_dict).encode()
    exception = KeycloakError(error_message=error_message, response_code=401)

    actual_result = keycloak_error_message(exception)

    assert actual_result == expected_result


def test_auth_keycloak_error_message_string():
    """Conversion from KeycloakError to error message should work with detail or not"""
    error_message = "Can't connect to server "
    exception = KeycloakError(error_message=error_message)
    assert keycloak_error_message(exception) == error_message
